#version 120

#define M_PI 3.1415926535897932384626433832795

uniform float u_shadowMaxAlpha, u_shadowGamma;
uniform vec3 u_projectedShadowDrawMin, u_projectedShadowDrawMax;
uniform vec2 u_xyDistanceMultiplier; // Usually 1.0, but can set to 0.0 to make an edge hard

varying vec3 v_worldPosition;
varying float v_lightRadius; // half penumbra width
varying vec2 v_unitTexCoord, v_constCastPlaneSize;
varying vec4 v_color;

void main()
{
    vec2 proportionFromEdge = vec2(0.5) - abs(vec2(0.5) - v_unitTexCoord); // this point distance from nearest corner, normalized so (0.5, 0.5) is center
    vec2 worldDistanceFromEdge = proportionFromEdge * v_constCastPlaneSize; // this point distance from nearest corner, in inches
    vec2 distanceIntoLight = max(vec2(2.0 * v_lightRadius) - worldDistanceFromEdge, vec2(0.0)) * u_xyDistanceMultiplier; // measured from "the edge" of the light shape, how much of it is visible to us at this point ({0, 0} in this case would be "all dark", {2*v_lightRadius,2*v_lightRadius} "all light")
    vec2 proportionIntoLight = distanceIntoLight / (v_lightRadius * 2.0);

#if 0 // DEBUG_RADIUS
    gl_FragColor = vec4(proportionIntoLight.x, proportionIntoLight.y, 0.0, 1.0) * u_shadowMaxAlpha;
    //gl_FragColor = vec4(v_unitTexCoord.s, v_unitTexCoord.t, 0.0, 1.0) * u_shadowMaxAlpha;
#else
    if (any(lessThan(v_worldPosition, u_projectedShadowDrawMin)) ||
        any(greaterThan(v_worldPosition, u_projectedShadowDrawMax))) {
#   if 0 // DEBUG_CLIPPING
        gl_FragColor = vec4(0.0, 0.0, 1.0, 1.0) * u_shadowMaxAlpha;
#   else
        discard;
#   endif
    } else {

#   if 0 // LIGHT_AS_SQUARE
        float lightEdge = v_lightRadius * 2.0;
        vec2 lightVector = vec2(lightEdge), zeroVector = vec2(0.0);
        distanceIntoLight /= lightVector;

        float visibleLight = (distanceIntoLight.x) + (1.0 - distanceIntoLight.x) * distanceIntoLight.y; // long segment plus short segment == the three possible areas of light added up
        if (visibleLight > 254.0/255.0) {
            //gl_FragColor = vec4(u_shadowMaxAlpha, 0.0, 0.0, u_shadowMaxAlpha);
            discard;
        } else {
            float fullLight = (1.0 * 1.0); // squared
            float redChannel = 0.0;

#       if 0 // DEBUG_PENUMBRA
            float lightCenterEdge = 0.05;
            vec2 distanceIntoCenterLight = min(max(distanceIntoLight - vec2(0.5 - lightCenterEdge / 2.0), zeroVector), lightCenterEdge) / lightCenterEdge;
            float centerLightWeight = 0.2;
            visibleLight += ((distanceIntoCenterLight.x) + (1.0 - distanceIntoCenterLight.x) * distanceIntoCenterLight.y) * centerLightWeight;
            redChannel += (distanceIntoCenterLight.x) + (1.0 - distanceIntoCenterLight.x) * distanceIntoCenterLight.y;
            fullLight += centerLightWeight;
#       endif

            float shadowAmount = u_shadowMaxAlpha * (1.0 - pow(clamp(visibleLight / fullLight, 0.0, 1.0), u_shadowGamma));
            gl_FragColor = vec4(redChannel, 0.0, 0.0, 1.0) * shadowAmount;
            //gl_FragColor = vec4(0.0, 0.0, 0.0, 1.0) * u_shadowMaxAlpha;
        }
        
#   else // LIGHT_AS_CIRCLE

        // TODO: We appear to have (lack of) perspective interpolation issues. v_lightRadius not interpolated correctly? Or are we doing something unstable in the vertex shader?
        vec2 distanceIntoLight = max(vec2(2.0 * v_lightRadius) - worldDistanceFromEdge, vec2(0.0)) * u_xyDistanceMultiplier;

#      if 0 // DEBUG_PENUMBRA
        gl_FragColor = vec4(vec2(distanceIntoLight / (v_lightRadius * 2.0)) * 0.6, 0.0, 1.0) * u_shadowMaxAlpha;
        if (any(lessThan(abs(distanceIntoLight - v_lightRadius), vec2(0.01))))
            gl_FragColor = vec4(0.0, 0.0, 0.8, 1.0) * u_shadowMaxAlpha;
        bvec2 lt = lessThan(distanceIntoLight, vec2(0.02)), gt = greaterThan(distanceIntoLight, vec2(0.005));
        if (any(bvec2(lt.x && gt.x, lt.y && gt.y))) // Surely there's something like: and(bvec2,bvec2)
            gl_FragColor = vec4(0.0, 0.0, 0.8, 1.0) * u_shadowMaxAlpha;
        return;
#      endif

        // Circle area approximation only works for up to a half-circle, so make sure we only have half!
        bvec2 visibleAreaMoreThanHalf = greaterThan(distanceIntoLight, vec2(v_lightRadius));
        distanceIntoLight = v_lightRadius - abs(distanceIntoLight - v_lightRadius);

        vec2 lightChordLength = 2.0 * sqrt(2.0 * v_lightRadius * distanceIntoLight - distanceIntoLight * distanceIntoLight);
        // Circle segment area approximation <http://mathworld.wolfram.com/CircularSegment.html>
        vec2 segmentArea = max(2.0/3.0 * lightChordLength * distanceIntoLight + pow(distanceIntoLight, vec2(3.0)) / (2.0 * lightChordLength), vec2(0.0));
        float doubleCountedArea = 0.5 * distanceIntoLight.x * distanceIntoLight.y; // Approximate double-counted bit with a triangle

        float fullLight = M_PI * v_lightRadius * v_lightRadius;
        float visibleLight = 0.0;
        visibleLight += visibleAreaMoreThanHalf.x ? (fullLight - segmentArea.x) : (segmentArea.x);
        visibleLight += visibleAreaMoreThanHalf.y ? (fullLight - segmentArea.y) : (segmentArea.y);
        if (all(visibleAreaMoreThanHalf))
            visibleLight += doubleCountedArea; // TODO: This never seems to be true. Why's that?
        else
            visibleLight -= doubleCountedArea;

        gl_FragColor = v_color * (u_shadowMaxAlpha * pow(clamp(1.0 - visibleLight / fullLight, 0.0, 1.0), u_shadowGamma));
#    endif
    }
#endif
}
